"""
A class encapsulating a Bismuth wallet (keys, address, crypto functions)

WIP
"""

# import base64
import json
# import logging

from os import path
from bismuthclient import bismuthcrypto
from Cryptodome.PublicKey import RSA
# from Cryptodome.Signature import PKCS1_v1_5
# from Cryptodome.Hash import SHA
# import getpass
# import hashlib

__version__ = '0.0.41'


class BismuthWallet():

    __slots__ = ('_address', '_wallet_file', 'verbose', '_encrypted', '_infos', "verbose", "key", "public_key")

    def __init__(self, wallet_file='wallet.der', verbose=False):
        self._wallet_file = None
        self._address = None
        self._infos = None
        self.key = None
        self.public_key = ''
        self.verbose = verbose
        self.load(wallet_file)

    def wallet_preview(self, wallet_file='wallet.der'):
        """
        Returns info about a wallet without actually loading it.
        """
        info = {'file': wallet_file, 'address': '', 'encrypted': False}
        try:
            with open(wallet_file, 'r') as f:
                content = json.load(f)
            info['address'] = content['Address']  # Warning case change!!!
            try:
                key = RSA.importKey(content['Private Key'])
                info['encrypted'] = False
            except:  # encrypted
                info['encrypted'] = True
        except:
            pass
        return info

    def info(self):
        """
        Returns a dict with info about the current wallet.
        :return:
        """
        return self._infos

    def load(self, wallet_file='wallet.der'):
        """
        Loads the wallet.der file

        :param wallet_file: string, a wallet file path
        """
        if self.verbose:
            print("Load", wallet_file)
        self._wallet_file = None
        self._address = None
        self._infos = {"address": '', 'file': wallet_file, 'encrypted': False}
        if not path.isfile(wallet_file):
            return

        self._wallet_file = wallet_file
        with open(wallet_file, 'r') as f:
            content = json.load(f)
            self._infos['address'] = content['Address']  # Warning case change!!!
            self._address = content['Address']
        try:
            self.key = RSA.importKey(content['Private Key'])
            self.public_key = content['Public Key']
        except:  # encrypted
            self._infos['encrypted'] = True

    def new(self, wallet_file='wallet.der'):
        """
        Creates a new wallet - only if the given file does not exist yet.
        Does not load it.

        :param wallet_file: string, a wallet file path
        :return True or False depending on the op success
        """
        if self.verbose:
            print("create", wallet_file)
        if path.isfile(wallet_file):
            if self.verbose:
                print("{} Already exists, canceled".format(wallet_file))
            return False
        bismuthcrypto.keys_new(wallet_file)
        return True


    @property
    def address(self):
        """Returns the currently loaded address, or None"""
        return self._address



"""
A class encapsulating a Bismuth wallet (keys, address, crypto functions)

This is a wallet.json file that can embed many different addresses/keys and is globally encrypted by a master password.

WIP
"""

from base64 import b64encode, b64decode
import json
import random
from copy import deepcopy

from os import path
from time import time
from bismuthclient import bismuthcrypto
from bismuthclient.simplecrypt import encrypt, decrypt
from Cryptodome.PublicKey import RSA
# from Cryptodome.Signature import PKCS1_v1_5
# from Cryptodome.Hash import SHA256
# import getpass
# import hashlib
import os, sys

__version__ = '0.0.2'


class BismuthMultiWallet():

    __slots__ = ('_address', '_wallet_file', 'verbose', '_infos', "verbose", "key", "public_key",
                 '_addresses', '_locked', '_data', '_master_password')

    def __init__(self, wallet_file: str='wallet.json', verbose: bool=False, seed: str=None):
        self._wallet_file = None
        self._address = None
        self._infos = None
        self.key = None
        self._data = None  # raw data
        self._locked = False
        self.public_key = ''
        self.verbose = verbose
        self._addresses = []  # clear data if unlocked
        self._master_password = ''
        self.load(wallet_file, seed=seed)

    def wallet_preview(self, wallet_file: str='wallet.der'):
        """
        Returns info about a LEGACY wallet without actually loading it.
        """
        """
        info = {'file': wallet_file, 'address': '', 'encrypted': False, 'count': 0}
        try:
            with open(wallet_file, 'r') as f:
                content = json.load(f)
            info['count'] = len(content['addresses'])
            info['encrypted'] = len(content['encrypted'])
        except:
            pass
        return info
        """
        info = {'file': wallet_file, 'address': '', 'encrypted': False}
        try:
            with open(wallet_file, 'r') as f:
                content = json.load(f)
            info['address'] = content['Address']  # Warning case change!!!
            try:
                key = RSA.importKey(content['Private Key'])
                info['encrypted'] = False
            except:  # encrypted
                info['encrypted'] = True
        except:
            pass
        return info

    def info(self):
        """
        Returns a dict with info about the current wallet.
        :return:
        """
        self._infos['count'] = len(self._data['addresses'])
        return self._infos

    def load(self, wallet_file: str='wallet.json', seed: str=None):
        """
        Loads the wallet.json file

        :param wallet_file: string, a wallet file path
        :param seed: None or string, an optional seed for reproducible tests. Do NOT use in prod.
        """
        if self.verbose:
            print("Load Multi", wallet_file)
        self._wallet_file = None
        self._address = None
        self._infos = {"address": '', 'file': wallet_file, 'encrypted': False}
        if seed:
            random.seed(seed)
        if not path.isfile(wallet_file):
            charset = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ&~#{([|-\_@)]=}+-*/<>!,;:.?%'
            salt = "".join(random.choice(charset) for x in range(random.randint(10, 20)))
            default = {"salt": salt, "spend": {"type": None, "value": None},
                       "version": __version__, "coin": "bis", "encrypted": False,
                       "addresses": []}
            with open(wallet_file, 'w') as f:
                json.dump(default, f)

        with open(wallet_file, 'r') as f:
            self._data = json.load(f)
            # No selected address by default.
            self._infos['address'] = ''
            self._address = ''
        self._infos['encrypted'] = self._data['encrypted']
        self._infos['spend'] = self._data['spend']
        self._wallet_file = wallet_file
        # If our wallet
        self._locked = self._data['encrypted']
        self._master_password = ''
        if not self._locked:
            self._addresses = deepcopy(self._data['addresses'])
            try:
                self._address = self._addresses[0]['address']
            except:
                self._address = None
        else:
            self._addresses = []
            self._address = None

    def save(self, wallet_file: str=None):
        if wallet_file is None:
            wallet_file = self._wallet_file
        with open(wallet_file, 'w') as f:
            json.dump(self._data, f)

    def password_ok(self, password: str):
        return password == self._master_password

    def encrypt(self, password:str='', current_password:str=None):
        """Encrypt - or re-encrypt """
        if len(self._addresses) <= 0:
            raise RuntimeWarning("Can't encrypt empty wallet.")
        if self._locked:
            raise RuntimeWarning("Can't encrypt locked down wallet.")
        if self._infos['encrypted']:
            # TODO
            raise RuntimeWarning("TODO: decrypt and re-encrypt - WIP")
        encrypted_addresses = []
        for address in self._addresses:
            content = json.dumps(address)
            encrypted_addresses.append(b64encode(encrypt(password, content)).decode('utf-8'))
        self._data['addresses'] = encrypted_addresses
        self._data['encrypted'] = True
        encrypted = b64encode(encrypt(self._master_password, self._data['spend'])).decode('utf-8')
        self._data['spend'] = encrypted
        self.save()
        self._master_password = password
        self._infos['encrypted'] = True
        self._locked = False

    def lock(self):
        """Lock the wallet"""
        if not self._data['encrypted']:
            raise RuntimeWarning("You have to encrypt your wallet first")
        if len(self._addresses) <= 0:
            raise RuntimeWarning("Can't lock empty wallet.")
        self._master_password = ''      # forget the pass
        self._locked = self._data['encrypted']
        if self._locked:
            # If wallet was encrypted, then forget the addresses also.
            self._addresses = []

    def unlock(self, password:str):
        """Sets the master password and unlocks the wallet"""
        if not self._locked:
            return
        if not self._infos['encrypted']:
            return
        # decrypt the addresses and keys
        try:
            addresses = []
            for address in self._data['addresses']:
                decoded = json.loads(decrypt(password, b64decode(address.encode('utf-8'))).decode('utf-8'))
                addresses.append(decoded)
            self._addresses = addresses
            self._master_password = password
            self._locked = False
            # Now decode the spend
            decoded = json.loads(decrypt(password, b64decode(self._data['spend'].encode('utf-8'))).decode('utf-8'))
            self._infos['spend'] = decoded
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_type, fname, exc_tb.tb_lineno)
            raise RuntimeWarning("Password does not seem to match")

    def new(self, wallet_file='wallet.der'):
        """
        Creates a new wallet - only if the given file does not exist yet.
        Does not load it.

        :param wallet_file: string, a wallet file path
        :return True or False depending on the op success
        """
        raise RuntimeWarning("Not to be used in multi context")
        """
        if self.verbose:
            print("create", wallet_file)
        if path.isfile(wallet_file):
            if self.verbose:
                print("{} Already exists, canceled".format(wallet_file))
            return False
        bismuthcrypto.keys_new(wallet_file)
        return True
        """

    def new_address(self, label: str='', password: str='', salt: str=''):
        """
        Add a new address to the wallet (and save it)
        """
        if self._infos['encrypted'] and self._locked:
            raise RuntimeError("Wallet must be unlocked")
        keys = bismuthcrypto.keys_gen(password=password, salt=salt)
        keys['label'] = label
        keys['timestamp'] = int(time())
        self._addresses.append(keys)
        if self._infos['encrypted']:
            content = json.dumps(keys)
            encrypted = b64encode(encrypt(self._master_password, content)).decode('utf-8')
            self._data['addresses'].append(encrypted)
        else:
            print('1')
            self._data['addresses'].append(keys)
        self.save()

    def set_spend(self, spend_type:str, spend_value: str, password: str=''):
        """Saves the spend protection if the pass is ok"""
        if not self.password_ok(password):
            raise RuntimeWarning("Password does not seem to match")
        if spend_type == 'None':
            spend_type = None
        spend = {'type': spend_type, 'value': spend_value}
        if self._infos['encrypted']:
            content = json.dumps(spend)
            encrypted = b64encode(encrypt(self._master_password, content)).decode('utf-8')
            self._data['spend'] = encrypted
        else:
            self._data['spend'] = spend
        self._infos['spend'] = spend
        self.save()

    def get_der_key(self, wallet_file: str='wallet.der', password: str=''):
        try:
            with open(wallet_file, 'r') as f:
                content = json.load(f)
                address = content['Address']  # Warning case change!!!
                if password:
                    content['Private Key'] = decrypt(password,b64decode(content['Private Key']))

                key = RSA.importKey(content['Private Key'])
                public_key = content['Public Key']
                private_key = content['Private Key']
                address = content["Address"]
                # TODO: check that address matches rebuilded pubkey
                return {"private_key": private_key, "public_key": public_key, "address": address, "label":'',
                        'timestamp': int(time())}
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_type, fname, exc_tb.tb_lineno)
            # encrypted
            return None

    def set_address(self, address: str=''):
        """Select an address from the wallet"""
        if self._infos['encrypted'] and self._locked:
            # TODO: check could be done via a decorator
            raise RuntimeError("Wallet must be unlocked")
        if not self.is_address_in_wallet(address):
            raise RuntimeError("Duplicate address")
        key = self.get_key(address)
        self.key = RSA.importKey(key['private_key'])
        self.public_key = key['public_key']
        self._address = address
        self._infos['address'] = address

    def is_address_in_wallet(self, address: str=''):
        if self._infos['encrypted'] and self._locked:
            # TODO: check could be done via a decorator
            raise RuntimeError("Wallet must be unlocked")
        for key in self._addresses:
            if address == key['address']:
                return True
        return False

    def get_key(self, address: str=''):
        if self._infos['encrypted'] and self._locked:
            # TODO: check could be done via a decorator
            raise RuntimeError("Wallet must be unlocked")
        for key in self._addresses:
            if address == key['address']:
                return key
        return None

    def import_der(self, wallet_der: str='wallet.der', label: str='', source_password: str=''):
        """Import an existing wallet.der like file into the wallet"""
        if self._infos['encrypted'] and self._locked:
            # TODO: check could be done via a decorator
            raise RuntimeError("Wallet must be unlocked")
        key = self.get_der_key(wallet_der, password=source_password)
        if not key:
            raise RuntimeWarning("Error importing the der file")
        key['label'] = label
        if self.is_address_in_wallet(key['address']):
            raise RuntimeError("Duplicate address")
        self._addresses.append(key)
        if self._infos['encrypted']:
            content = json.dumps(key)
            encrypted = b64encode(encrypt(self._master_password, content)).decode('utf-8')
            self._data['addresses'].append(encrypted)
        else:
            print('1')
            self._data['addresses'].append(key)
        self.save()

    @property
    def address(self):
        """Returns the currently loaded address, or None"""
        return self._address



"""
Wallet server benchmarking functions - ported from gui wallet, converted to use the native api class.
"""


import collections
import logging
import socket
import time

from bismuthclient import rpcconnections

DEFAULT_PORT = 5658


def convert_ip_port(ip, some_port):
    """
    Helper: Get ip and port, but extract port from ip if ip was as ip:port

    :param ip:
    :param some_port: default port
    :return: (ip, port)
    """
    if ':' in ip:
        ip, some_port = ip.split(':')
    return ip, some_port


def connectible(ipport):
    """
    Helper: return True if the ip:port can be connected to, without sending any command.
    """
    try:
        s = socket.socket()
        s.settimeout(3)
        ip, local_port = convert_ip_port(ipport, DEFAULT_PORT)
        s.connect((ip, int(local_port)))
        return True
    except Exception as e:
        print(str(e))
        return False


def time_measure(light_ip_list, app_log=None):
    """
    Measure answer time for the servers in light_ip list

    :param light_ip_list: a list of "ip:port" wallet servers
    :param app_log: an optional logger
    :return: A sorted light_ip list
    """
    if not app_log:
        app_log = logging

    port = DEFAULT_PORT
    result_collection = {ip: [0, 0] for ip in light_ip_list}

    rpcconnections.LTIMEOUT = 3

    for address in result_collection:
        try:
            ip, local_port = convert_ip_port(address, port)
            app_log.info("Attempting to benchmark {}:{}".format(ip, local_port))

            client = rpcconnections.Connection("{}:{}".format(ip, local_port))

            if local_port == DEFAULT_PORT: #doesn't work if a node uses non standard port, bench in else-path - will fail
                #start benchmark
                timer_start = time.time()
                result = client.command("statusget")
                timer_result = (time.time() - timer_start) * 5 #penalty to prio Wallet-Servers before nodes. local node should be so fast, to be still fastest, else it is better that a wallet-server is chosen!
                result_collection[address] = timer_result, result[8][7]
                app_log.warning("Result for {}:{}, a normal node, penalty-factor *5 (real result time/5): {}".format(ip, local_port, timer_result))
                #finish benchmark
            else:
                #start benchmark
                timer_start = time.time()
                result = client.command("statusget")
                result_ws = client.command("wstatusget")
                timer_result = time.time() - timer_start
                #finish benchmark and load balance if too many clients
                ws_clients = result_ws.get('clients')
                if ws_clients > 300:
                    timer_result = timer_result + ws_clients/1000
                    app_log.warning("Result for {}:{}, modified due to high client load: {}".format(ip, local_port, timer_result))
                elif ws_clients > 150:
                    timer_result = timer_result + ws_clients/10000
                    app_log.warning("Result for {}:{}, modified due to client load: {}".format(ip, local_port, timer_result))
                else:
                    app_log.warning("Result for {}:{}, low load - unmodified: {}".format(ip, local_port, timer_result))
                result_collection[address] = timer_result, result[8][7]

        except Exception as e:
            app_log.warning("Cannot benchmark {}:{}".format(ip, local_port))

    # sort IPs for measured Time
    bench_result = collections.OrderedDict(sorted((value[0], key) for (key, value) in result_collection.items()))
    light_ip = list(bench_result.values())

    max_height_temp = list(result_collection.values())
    max_height = max(list(zip(*max_height_temp))[1])
    for key, value in result_collection.items():
        if int(value[1]) < (max_height - 5):
            try:
                light_ip.remove(key)
                light_ip.append(key)
            except Exception as e:
                pass
    return light_ip



"""
Wrapper around the official Bismuth API from api.bismuth.live

WIP
"""

import logging
import requests
import time

# Light wallet benchmark and helpers
from bismuthclient import lwbench
from distutils.version import LooseVersion

__version__ = '0.0.2'


def get_wallet_servers_legacy(light_ip_list='', app_log=None, minver='0'):
    """
    Use different methods to return the best possible list of wallet servers,
    sorted
    ip:port"""

    if not app_log:
        app_log = logging

    while True:
        # If we have 127.0.0.1 in the list, first try it
        if '127.0.0.1:5658' in light_ip_list or '127.0.0.1' in light_ip_list:
            if lwbench.connectible('127.0.0.1:5658'):
                # No need to go further.
                return ['127.0.0.1:5658']

        # Then try the new API
        wallets = []
        try:
            rep = requests.get("http://api.bismuth.live/servers/wallet/legacy.json")
            if rep.status_code == 200:
                wallets = rep.json()
        except Exception as e:
            app_log.warning("Error {} getting Server list from API, using lwbench instead".format(e))

        if not wallets:
            # no help from api, use previous benchmark
            ipport_list = lwbench.time_measure(light_ip_list, app_log)
            return ipport_list

        # We have a server list, order by load
        sorted_wallets = sorted([wallet for wallet in wallets if wallet['active'] and
                                 wallet.get('version', '0') >= LooseVersion(minver)],
                                key=lambda k: (k['clients']+1)/(k['total_slots']+2))

        if sorted_wallets:
            return ["{}:{}".format(wallet['ip'], wallet['port']) for wallet in sorted_wallets]

        # If we get here, all hope is lost!
        app_log.warning("No connectible server... let try again in a few sec")
        time.sleep(10)



"""
Bismuth default/legacy connection layer.
Json over sockets
This file is no more compatible with the Bismuth code, it's been converted to a class
EggPool 2018
"""

import json
import socket
import time
import threading

# Logical timeout
LTIMEOUT = 45
# Fixed header length
SLEN = 10


__version__ = '0.1.10'


class Connection(object):
    """Connection to a Bismuth Node. Handles auto reconnect when needed"""

    __slots__ = ('ipport', 'verbose', 'sdef', 'stats', 'last_activity', 'command_lock', 'raw')

    def __init__(self, ipport, verbose=False, raw=False):
        """ipport is an (ip, port) tuple"""
        self.ipport = ipport
        if ':' in ipport:
            ip, port = ipport.split(':')
            self.ipport = (ip, int(port))
        self.verbose = verbose
        self.raw = raw
        self.sdef = None
        self.last_activity = 0
        self.command_lock = threading.Lock()
        self.check_connection()

    def check_connection(self):
        """Check connection state and reconnect if needed."""
        if not self.sdef:
            try:
                if self.verbose:
                    print("Connecting to", self.ipport)
                self.sdef = socket.socket()
                self.sdef.settimeout(LTIMEOUT)
                self.sdef.connect(self.ipport)
                self.last_activity = time.time()
            except Exception as e:
                self.close()
                raise RuntimeError("Connections: {}".format(e))

    def _send(self, data, slen=SLEN, retry=True):
        """Sends something to the server"""
        self.check_connection()
        try:
            self.sdef.settimeout(LTIMEOUT)
            # Make sure the packet is sent in one call
            sdata = str(json.dumps(data))
            res = self.sdef.sendall(str(len(sdata)).encode("utf-8").zfill(slen)+sdata.encode("utf-8"))
            if self.raw:
                print("sending raw:")
                print(str(len(sdata)).encode("utf-8").zfill(slen)+sdata.encode("utf-8"))
            self.last_activity = time.time()
            # res is always 0 on linux
            if self.verbose:
                print("send ", data)
            return True
        except Exception as e:
            # send failed, try to reconnect
            # TODO: handle tries #
            self.close()
            if retry:
                if self.verbose:
                    print("Send failed ({}), trying to reconnect".format(e))
                self.check_connection()
            else:
                if self.verbose:
                    print("Send failed ({}), not retrying.".format(e))
                return False
            try:
                self.sdef.settimeout(LTIMEOUT)
                # Make sure the packet is sent in one call
                self.sdef.sendall(str(len(str(json.dumps(data)))).encode("utf-8").zfill(slen)+str(json.dumps(data)).encode("utf-8"))
                return True
            except Exception as e:
                self.close()
                raise RuntimeError("Connections: {}".format(e))

    def _receive(self, slen=SLEN):
        """Wait for an answer, for LTIMEOUT sec."""
        self.check_connection()
        self.sdef.settimeout(LTIMEOUT)
        if self.raw:
            print("getting raw:")
        try:
            data = self.sdef.recv(slen)
            if self.raw:
                raw = data
            if not data:
                self.close()
                raise RuntimeError("Socket EOF")
            data = int(data)  # receive length
        except socket.timeout as e:
            self.close()
            return ""
        try:
            chunks = []
            bytes_recd = 0
            while bytes_recd < data:
                chunk = self.sdef.recv(min(data - bytes_recd, 2048))
                if not chunk:
                    raise RuntimeError("Socket EOF2")
                chunks.append(chunk)
                bytes_recd = bytes_recd + len(chunk)
            self.last_activity = time.time()
            if self.raw:
                print(raw + b''.join(chunks))
            segments = b''.join(chunks).decode("utf-8")
            return json.loads(segments)
        except Exception as e:
            """
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_type, fname, exc_tb.tb_lineno)
            """
            self.close()
            raise RuntimeError("Connections: {}".format(e))

    def command(self, command, options=None):
        """
        Sends a command and return its raw result.
        options has to be a list.
        Each item of options will be sent separately. So If you ant to send a list, pass a list of list.
        """
        with self.command_lock:
            try:
                self._send(command)
                if options:
                    for option in options:
                        self._send(option, retry=False)
                ret = self._receive()
                return ret
            except Exception as e:
                """
                exc_type, exc_obj, exc_tb = sys.exc_info()
                fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
                print(exc_type, fname, exc_tb.tb_lineno)
                """
                # TODO : better handling of tries and delay between
                if self.verbose:
                    print("Error <{}> sending command, trying to reconnect.".format(e))
                self.close()
                self.check_connection()
                self._send(command)
                if options:
                    for option in options:
                        self._send(option, retry=False)
                ret = self._receive()
                return ret

    def close(self):
        """Close the socket"""
        if self.sdef:
            try:
                self.sdef.close()
            except Exception as e:
                pass
            finally:
                self.sdef = None


if __name__ == "__main__":
    print("I'm a module, can't run!")



"""
A all in one Bismuth Native client that connects to local or distant wallet servers
"""

import base64
# import json
import logging
import time
import sys
import os

# from bismuthclient import async_client
from bismuthclient import bismuthapi
from bismuthclient.bismuthwallet import BismuthWallet
from bismuthclient.bismuthmultiwallet import BismuthMultiWallet
from bismuthclient import bismuthcrypto
from bismuthclient import rpcconnections
from bismuthclient import lwbench
from bismuthclient.bismuthformat import TxFormatter, AmountFormatter
from os import path, scandir

__version__ = '0.0.49'


class BismuthClient():

    __slots__ = ('initial_servers_list', 'servers_list', 'app_log', '_loop', 'address', '_current_server',
                 'wallet_file', '_wallet', '_connection', '_cache', 'verbose')

    def __init__(self, servers_list=None, app_log=None, loop=None, wallet_file='', verbose=False):
        """
        Init the main class.

        :param servers_list: list of "ip:port" wallet servers
        :param app_log:
        :param loop: None of an asyncio loop
        :param address:
        """
        self.verbose = verbose
        if not servers_list:
            servers_list = []
        self.initial_servers_list = servers_list
        if app_log:
            self.app_log = app_log
        elif logging.getLogger("tornado.application"):
            self.app_log = logging.getLogger("tornado.application")
        else:
            self.app_log = logging
        self._loop = loop
        self.wallet_file = None
        self._wallet = None
        self.address = None
        self.load_wallet(wallet_file)
        self.servers_list = servers_list
        self._current_server = None
        self._connection = None
        self._cache = {}

    def _get_cached(self, key, timeout_sec=30):
        if key in self._cache:
            data = self._cache[key]
            if data[0] + timeout_sec >= time.time():
                """                
                if self.verbose:
                    self.app_log.info("Cache Hit on {}".format(key))
                    # print(data[1])
                """
                return data[1]
        return None

    def _set_cache(self, key, value):
        self._cache[key] = (time.time(), value)

    def clear_cache(self):
        self._cache = {}

    @property
    def current_server(self):
        return self._current_server

    @staticmethod
    def user_subdir(subdir):
        """Returns a path to subdir in the user data directory. Path will be created if it does not exist."""
        home = os.path.expanduser('~')
        location = os.path.join(home, subdir)
        if not os.path.isdir(location):
            os.makedirs(location, exist_ok=True)
        return location

    def list_wallets(self, scan_dir='wallets'):
        """
        Returns a list of dict for each wallet file found in the dir to scan.

        Each dict has the following keys: 'file', 'address', 'encrypted'

        :param scan_dir: string, the dir to scan for wallet (*.der files).
        """
        wallets = []
        for entry in scandir(scan_dir):
            print(entry)
            if entry.name.endswith('.der') and entry.is_file():
                 wallets.append(self._wallet.wallet_preview(entry.path))
        # TODO: sorts by name
        return wallets

    def latest_transactions(self, num=10, for_display=False):
        """
        Returns the list of the latest num transactions for the current address.

        Each transaction is a dict with the following keys:
        `["block_height", "timestamp", "address", "recipient", "amount", "signature", "public_key", "block_hash", "fee", "reward", "operation", "openfield"]`
        """
        if not self.address or not self._wallet:
            return []
        try:
            key = "tx{}".format(num)
            cached = self._get_cached(key)
            if cached:
                return cached
            transactions = self.command("addlistlim", [self.address, num])
        except:
            # TODO: Handle retry, at least error message.
            transactions = []

        #json = [dict(zip(["block_height", "timestamp", "address", "recipient", "amount", "signature", "public_key", "block_hash", "fee", "reward", "operation", "openfield"], tx)) for tx in transactions]
        json = [TxFormatter(tx).to_json(for_display=for_display) for tx in transactions]
        # print(json)
        self._set_cache(key, json)
        return json

    def balance(self, for_display=False):
        """
        Returns the current balance for the current address.
        """
        if not self.address or not self._wallet:
            return 'N/A'
        try:
            cached = self._get_cached('balance')
            if cached:
                return cached
            balance = self.command("balanceget", [self.address])
            balance = balance[0]
            self._set_cache('balance', balance)
        except:
            # TODO: Handle retry, at least error message.
            balance = 'N/A'
        if for_display:
            balance = AmountFormatter(balance).to_string(leading=0)
        if balance == '0E-8':
            balance = 0.000
        return balance

    def send(self, recipient: str, amount: float, operation: str='', data: str=''):
        """
        Sends the given tx
        """
        try:
            timestamp = time.time()
            public_key_hashed = base64.b64encode(self._wallet.public_key.encode('utf-8'))
            signature_enc = bismuthcrypto.sign_with_key(timestamp, self.address, recipient, amount, operation, data, self._wallet.key)
            txid = signature_enc[:56]
            tx_submit = ( '%.2f' % timestamp, self.address, recipient, '%.8f' % float(amount),
                          str(signature_enc), str(public_key_hashed.decode("utf-8")), operation, data)
            reply = self.command('mpinsert', [tx_submit])
            if self.verbose:
                print("Server replied '{}'".format(reply))
            if reply[-1] != "Success":
                print("Error '{}'".format(reply))
                return None
            if not reply:
                print("Server timeout")
                return None
            return txid
        except Exception as e:
            print(str(e))
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_type, fname, exc_tb.tb_lineno)

    def sign(self, message: str):
        """
        Signs the given message
        """
        try:
            signature = bismuthcrypto.sign_message_with_key(message, self._wallet.key)
            return signature
        except Exception as e:
            print(str(e))
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_type, fname, exc_tb.tb_lineno)
            raise


    def status(self):
        """
        Returns the current status of the wallet server
        """
        try:
            cached = self._get_cached('status')
            if cached:
                return cached
            status = self.command("statusjson")
            try:
                status['extended'] = self.command("wstatusget")
            except:
                status['extended'] = None
            self._set_cache('status', status)
        except:
            # TODO: Handle retry, at least error message.
            status = {}
        return status

    def load_wallet(self, wallet_file='wallet.der'):
        """
        Tries to load the wallet file

        :param wallet_file: string, a wallet.der file
        """
        # Default values, fail
        self.wallet_file = None
        self.address = None
        self._wallet = None
        self._wallet = BismuthWallet(wallet_file, verbose=self.verbose)
        self.wallet_file = wallet_file
        if self.address != self._wallet.address:
            self.clear_cache()
        self.address = self._wallet.address

    def load_multi_wallet(self, wallet_file='wallet.json'):
        """
        Tries to load the wallet file

        :param wallet_file: string, a wallet.json file
        """
        # TODO: Refactor
        self.wallet_file = None
        self.address = None
        self._wallet = None
        self._wallet = BismuthMultiWallet(wallet_file, verbose=self.verbose)
        if len(self._wallet._data["addresses"]) == 0:
            # Create a first address by default
            self._wallet.new_address(label="default")
        self.wallet_file = wallet_file
        if self.address != self._wallet.address:
            self.clear_cache()
        self.address = self._wallet.address

    def set_address(self, address: str=''):
        if not type(self._wallet) == BismuthMultiWallet:
            raise RuntimeWarning("Not a Multiwallet")
        self._wallet.set_address(address)
        if self.address != self._wallet.address:
            self.clear_cache()
        self.address = self._wallet.address

    def new_wallet(self, wallet_file='wallet.der'):
        """
        Creates a new wallet if it does not already exists

        :param wallet_file: string, a wallet.der file
        """
        # Default values, fail
        wallet = BismuthWallet(wallet_file, verbose=self.verbose)
        return wallet.new(wallet_file)


    def wallet(self, full=False):
        """
        returns info about the currently loaded wallet

        if full is True, also force a check of the current balance.
        """
        return self._wallet.info()

    def info(self):
        """
        returns a dict with server info: ip, port, latest server status
        """
        connected = False
        if self._connection:
            connected = bool(self._connection.sdef)
        info = {"wallet": self.wallet_file, "address": self.address, "server": self._current_server, "servers_list": self.servers_list, "connected": connected}
        return info

    def get_server(self):
        """
        Tries to find the best available server given the config and sets self._current_server for later use.

        Returns the first connectible server.
        """
        # Use the API or bench to get the best one.
        if not len(self.initial_servers_list):
            self.servers_list = bismuthapi.get_wallet_servers_legacy(self.initial_servers_list, self.app_log, minver='0.1.5')
        else:
            self.servers_list = self.initial_servers_list
        # Now try to connect
        if self.verbose:
            print("self.servers_list", self.servers_list)
        for server in self.servers_list:
            if self.verbose:
                print("test server", server)
            if lwbench.connectible(server):
                self._current_server = server
                # TODO: if self._loop, use async version
                if self.verbose:
                    print("connect server", server)
                self._connection = rpcconnections.Connection(server, verbose=self.verbose)
                return server
        self._current_server = None
        self._connection = None
        # TODO: raise
        return None

    def command(self, command, options=None):
        """
        Makes sure we have a connection, runs a command and sends back the result.

        :param command: the command as a string
        :param options: optional options to the command, as a list if needed
        :return: the result as a native structure
        """
        if not self._current_server:
            # TODO: failsafe if can't connect
            self.get_server()
        if self.verbose:
            print("command {}, {}".format(command, options))
        return self._connection.command(command, options)



"""
Crypto related functions for Bismuth.

Keys, wallet loading, signature

Legacy functions kept for compatibility.
See new BismuthWallet Class
"""

import base64
import getpass
import hashlib
import json
import os
import re

from Cryptodome.Hash import SHA
from Cryptodome.PublicKey import RSA
from Cryptodome.Signature import PKCS1_v1_5
from Cryptodome.Protocol.KDF import PBKDF2
from os import path
from bismuthclient.simplecrypt import *


__version__ = '0.0.23'


def sign_rsa(timestamp, address, recipient, amount, operation, openfield, key, public_key_hashed):
    if not key:
        raise BaseException("The wallet is locked, you need to provide a decrypted key")

    transaction = (str (timestamp), str (address), str (recipient), '%.8f' % float (amount), str (operation), str (openfield))  # this is signed, float kept for compatibility

    h = SHA.new (str(transaction).encode())
    signer = PKCS1_v1_5.new (key)
    signature = signer.sign (h)
    signature_enc = base64.b64encode(signature)

    verifier = PKCS1_v1_5.new (key)
    if verifier.verify (h, signature):
        return_value = str (timestamp), str (address), str (recipient), '%.8f' % float (amount), str (signature_enc.decode ("utf-8")), str (public_key_hashed.decode ("utf-8")), str (operation), str (openfield)  # float kept for compatibility
    else:
        return_value = False

    return return_value


def format_transaction(timestamp: float, address: str, recipient: str, amount: float, operation: str, openfield: str):
    """
    Returns the formatted tuple to use as transaction part and to be signed
    This exact formatting is MANDATORY - We sign a char buffer where every char counts.
    """
    str_timestamp = '%.2f' % timestamp
    str_amount = '%.8f' % amount
    transaction = (str_timestamp, address, recipient, str_amount, operation, openfield)
    return transaction


def stringify_transaction(timestamp: float, address: str, recipient: str, amount: float, operation: str, openfield: str):
    """Formats the transaction items into the string buffer to be signed"""
    transaction = format_transaction(timestamp, address, recipient, amount, operation, openfield)
    return str(transaction).encode("utf-8")


def sign_with_key(timestamp: float, address: str, recipient: str, amount: float, operation: str, openfield: str, key):
    # Sign with key - This is a helper function
    # Returns the encoded sig as a string
    as_string = stringify_transaction(timestamp, address, recipient, amount, operation, openfield)
    # print("As String1", as_string)
    h = SHA.new(as_string)
    signer = PKCS1_v1_5.new(key)
    signature = signer.sign(h)
    # print("signature1", signature)
    signature_enc = base64.b64encode(signature)
    verifier = PKCS1_v1_5.new(key)
    if verifier.verify(h, signature):
        print("OK")
        return signature_enc.decode("utf-8")
    else:
        return False


def sign_message_with_key(message: str, key):
    # Sign with key - This is a helper function
    # Returns the b64 encoded sig as a string
    h = SHA.new(message.encode('utf-8'))
    signer = PKCS1_v1_5.new(key)
    signature = signer.sign(h)
    signature_enc = base64.b64encode(signature)
    verifier = PKCS1_v1_5.new(key)
    if verifier.verify(h, signature):
        # print("OK")
        return signature_enc.decode("utf-8")
    else:
        return False


def keys_check(app_log, keyfile):
    # key maintenance
    if os.path.isfile("privkey.der") is True:
        app_log.warning("privkey.der found")
    elif os.path.isfile("privkey_encrypted.der") is True:
        app_log.warning("privkey_encrypted.der found")
        os.rename("privkey_encrypted.der","privkey.der")

    elif os.path.isfile (keyfile) is True:
        app_log.warning ("{} found".format(keyfile))
    else:
        # generate key pair and an address

        key = RSA.generate(4096)
        #public_key = key.publickey()

        private_key_readable = key.exportKey().decode("utf-8")
        public_key_readable = key.publickey().exportKey().decode("utf-8")
        address = hashlib.sha224(public_key_readable.encode("utf-8")).hexdigest()  # hashed public key
        # generate key pair and an address

        app_log.info("Your address: {}".format(address))
        app_log.info("Your public key: {}".format(public_key_readable))

        # export to single file
        keys_save(private_key_readable, public_key_readable, address, keyfile)
        # export to single file


def keys_new(keyfile):
    if path.isfile(keyfile):
        return False
    key = RSA.generate(4096)
    # public_key = key.publickey()
    private_key_readable = key.exportKey().decode("utf-8")
    public_key_readable = key.publickey().exportKey().decode("utf-8")
    address = hashlib.sha224(public_key_readable.encode("utf-8")).hexdigest()  # hashed public key
    # export to single file
    keys_save(private_key_readable, public_key_readable, address, keyfile)
    return True


def keys_gen(password: str='', salt:str='', count=10000, verbose=False):
    """
    Optionnally deterministic RSA Key generation from password and salt. To be used by tests only atm.
    :param password:
    :param salt:
    :param count:
    :param verbose:
    :return:
    """
    if password:
        def my_rand(n):
            # kluge: use PBKDF2 with count=1 and incrementing salt as deterministic PRNG
            my_rand.counter += 1
            return PBKDF2(master_key, "my_rand:%d" % my_rand.counter, dkLen=n, count=1)
        if verbose:
            print("Generating master key")
        master_key = PBKDF2(password, salt, count=count)
        my_rand.counter = 0
        if verbose:
            print("Generating RSA key")
        key = RSA.generate(4096, randfunc=my_rand)
    else:
        if verbose:
            print("Generating RSA key")
        key = RSA.generate(4096)
    # public_key = key.publickey()
    private_key_readable = key.exportKey().decode("utf-8")
    public_key_readable = key.publickey().exportKey().decode("utf-8")
    address = hashlib.sha224(public_key_readable.encode("utf-8")).hexdigest()  # hashed public key
    # export to single file
    return {'private_key': private_key_readable, 'public_key': public_key_readable, 'address': address}


def keys_save(private_key_readable, public_key_readable, address, file):
    wallet_dict = {}
    wallet_dict['Private Key'] = private_key_readable
    wallet_dict['Public Key'] = public_key_readable
    wallet_dict['Address'] = address

    if not isinstance(file,str):
        file = file.name

    with open (file, 'w') as keyfile:
        json.dump (wallet_dict, keyfile)


def keys_load(privkey="privkey.der", pubkey="pubkey.der"):
    keyfile = "wallet.der"
    if os.path.exists("wallet.der"):
        print("Using modern wallet method")
        return keys_load_new ("wallet.der")

    else:
        # print ("loaded",privkey, pubkey)
        # import keys
        try:  # unencrypted
            key = RSA.importKey(open(privkey).read())
            private_key_readable = key.exportKey ().decode ("utf-8")
            # public_key = key.publickey()
            encrypted = False
            unlocked = True

        except:  # encrypted
            encrypted = True
            unlocked = False
            key = None
            private_key_readable = open(privkey).read()

        # public_key_readable = str(key.publickey().exportKey())
        public_key_readable = open(pubkey.encode('utf-8')).read()

        if (len(public_key_readable)) != 271 and (len(public_key_readable)) != 799:
            raise ValueError("Invalid public key length: {}".format(len(public_key_readable)))

        public_key_hashed = base64.b64encode(public_key_readable.encode('utf-8'))
        address = hashlib.sha224(public_key_readable.encode('utf-8')).hexdigest()

        print("Upgrading wallet")
        keys_save (private_key_readable, public_key_readable, address, keyfile)

        return key, public_key_readable, private_key_readable, encrypted, unlocked, public_key_hashed, address, keyfile


def keys_unlock(private_key_encrypted):
    password = getpass.getpass ()
    encrypted_privkey = private_key_encrypted
    decrypted_privkey = decrypt(password, base64.b64decode (encrypted_privkey))
    key = RSA.importKey (decrypted_privkey)  # be able to sign
    private_key_readable = key.exportKey ().decode ("utf-8")
    return key, private_key_readable


def keys_load_new(keyfile="wallet.der"):
    # import keys
    with open (keyfile, 'r') as keyfile:
        wallet_dict = json.load (keyfile)

    private_key_readable = wallet_dict['Private Key']
    public_key_readable = wallet_dict['Public Key']
    address = wallet_dict['Address']

    try:  # unencrypted
        key = RSA.importKey(private_key_readable)
        encrypted = False
        unlocked = True

    except:  # encrypted
        encrypted = True
        unlocked = False
        key = None

    # public_key_readable = str(key.publickey().exportKey())
    if (len(public_key_readable)) != 271 and (len(public_key_readable)) != 799:
        raise ValueError("Invalid public key length: {}".format(len(public_key_readable)))

    public_key_hashed = base64.b64encode(public_key_readable.encode('utf-8'))

    return key, public_key_readable, private_key_readable, encrypted, unlocked, public_key_hashed, address, keyfile


# Dup code, not pretty, but would need address module to avoid dup
def address_validate(address):
    return re.match('[abcdef0123456789]{56}', address)


# Dup code, not pretty, but would need address module to avoid dup
def validate_pem(public_key):
    # verify pem as cryptodome does
    pem_data = base64.b64decode(public_key).decode("utf-8")
    regex = re.compile("\s*-----BEGIN (.*)-----\s+")
    match = regex.match(pem_data)
    if not match:
        raise ValueError("Not a valid PEM pre boundary")
    marker = match.group(1)
    regex = re.compile("-----END (.*)-----\s*$")
    match = regex.search(pem_data)
    if not match or match.group(1) != marker:
        raise ValueError("Not a valid PEM post boundary")
        # verify pem as cryptodome does



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `bismuthclient` package.


Warning: NEVER use the test addresses or seed, as they are public!
"""

import pytest
import sys
import os

sys.path.append('../')
from bismuthclient.bismuthmultiwallet import BismuthMultiWallet
from bismuthclient.bismuthcrypto import keys_gen


def test0_ainit():
    if not os.path.exists("tmp"):
        os.mkdir("tmp")
    try:
        os.remove("tmp/wallet1.json")
    except:
        pass
    try:
        os.remove("tmp/wallet2.json")
    except:
        pass
    try:
        os.remove("tmp/wallet3.json")
    except:
        pass


def test1_create_clear_multiwallet():
    wallet = BismuthMultiWallet("tmp/wallet1.json", seed="seed1")


def test2_add_address():
    wallet = BismuthMultiWallet("tmp/wallet1.json", seed="seed2")
    wallet.new_address(label="ad1", password="1", salt="1")
    # wallet.new_address(label="ad2", password="2", salt="1")


def test2b_read_address():
    wallet = BismuthMultiWallet("tmp/wallet1.json")
    assert wallet._addresses[0]['address'] == '59994eac4a36942fdbb05e33a353f72850f26112123840844bedb870'


def test3_create_empty_encrypted_multiwallet():
    wallet = BismuthMultiWallet("tmp/wallet2.json", seed="seed2")
    with pytest.raises(Exception) as e_info:
        wallet.encrypt("some_password", current_password=None)


def test4a_create_encrypted_multiwallet():
    wallet = BismuthMultiWallet("tmp/wallet3.json", seed="seed2")
    wallet.new_address(label="ad1", password="1", salt="1")  # Will auto save
    wallet.encrypt("some_password", current_password=None)  # Will auto save
    wallet.lock()
    wallet.unlock("some_password")
    assert wallet._addresses[0]['address'] == '59994eac4a36942fdbb05e33a353f72850f26112123840844bedb870'


def test4b_addto_encrypted_multiwallet():
    wallet = BismuthMultiWallet("tmp/wallet3.json")
    wallet.unlock("some_password")
    wallet.new_address(label="ad2", password="2", salt="2")  # Will auto save
    for address in wallet._addresses:
        print(address['address'])


if __name__ == "__main__":
    # print(keys_gen('test', 'salt', verbose=True))
    test0_ainit()
    # test2_add_address()
    # test2b_read_address()
    test4a_create_encrypted_multiwallet()
    test4b_addto_encrypted_multiwallet()



"""
Generic helpers  Bismuth
"""

import re
import hashlib
import base64

__version__ = '0.0.2'



def checksum(string):
    """ Base 64 checksum of MD5. Used by bisurl"""
    m = hashlib.md5()
    m.update(string.encode("utf-8"))
    return base64.b85encode(m.digest()).decode("utf-8")


class BismuthUtil():
    """Static helper utils"""


    @staticmethod
    def valid_address(address: str):
        """Says if that address looks ok"""
        return re.match('[abcdef0123456789]{56}', address)

    @staticmethod
    def fee_for_tx(message=''):
        """Returns fee for the matching message"""
        return 0.01 + len(message) / 100000

    @staticmethod
    def height_to_supply(height):
        """Gives total supply at a given block height"""
        R0 = 11680000.4
        delta = 2e-6
        pos = 0.8
        pow = 12.6
        N = height - 8e5
        dev_rew = 1.1
        R = dev_rew * R0 + N * (pos + dev_rew * (pow - N / 2 * delta))
        return R

    @staticmethod
    def create_bis_url(recipient, amount, operation, openfield):
        """
        Constructs a bis url from tx elements
        """

        # Only command supported so far.
        command = "pay"
        openfield_b85_encode = base64.b85encode(openfield.encode("utf-8")).decode("utf-8")
        operation_b85_encode = base64.b85encode(operation.encode("utf-8")).decode("utf-8")
        url_partial = "bis://{}/{}/{}/{}/{}/".format(command,recipient,amount,operation_b85_encode,openfield_b85_encode)
        url_constructed = url_partial + checksum(url_partial)
        return url_constructed

    @staticmethod
    def read_url(url):
        """
        Takes a bis url, checks its checksum and gives the components
        """
        url_split = url.split("/")
        reconstruct = "bis://{}/{}/{}/{}/{}/".format(url_split[2],url_split[3],url_split[4],url_split[5],url_split[6],url_split[7])
        operation_b85_decode = base64.b85decode(url_split[5]).decode("utf-8")
        openfield_b85_decode = base64.b85decode(url_split[6]).decode("utf-8")

        if checksum(reconstruct) == url_split[7]:
            url_deconstructed = {"recipient": url_split[3], "amount": url_split[4], "operation": operation_b85_decode,
                                 "openfield": openfield_b85_decode}
            return url_deconstructed
        else:
            return {'Error': 'Checksum failed'}



"""
Formatting helpers and converters for Bismuth
"""

import datetime
# import re

__version__ = '0.0.32'


# TODO: common ancestor, factorize

class TxFormatter():
    """Formats a Bismuth Transaction"""

    __slots__ = ['tx']
    _tx_keys = ["block_height", "timestamp", "address", "recipient", "amount", "signature", "public_key", "block_hash",
                "fee", "reward", "operation", "openfield"]

    def __init__(self, tx):
        self.tx = tx

    def to_string(self):
        return str(self.tx)

    def to_json(self, for_display=False):
        json = dict(zip(self._tx_keys, self.tx))
        if for_display:
            json['datetime'] = datetime.datetime.utcfromtimestamp(json['timestamp']).strftime('%Y-%m-%d %H:%M:%S')
            json['amount'] = AmountFormatter(json['amount']).to_string()
            json['fee'] = AmountFormatter(json['fee']).to_string(leading=1)
            if json['operation'] == '0':
                json['operation'] = ''
        return json


class AmountFormatter():
    """Format a BIS amount to string"""

    __slots__ = ['amount']

    def __init__(self, amount):
        self.amount = float(amount)

    def to_string(self, decimals=3, leading=0):
        temp = round(self.amount, decimals)
        int_part, decimal_part = str(temp).split('.')
        while len(decimal_part) < decimals:
            decimal_part += '0'
        out = "{integer: >{fill}}.{decimal}".format(integer=int_part, decimal=decimal_part, fill=leading)
        return out



"""
Async TCP Client for Bismuth Wallet

Eggdrasyl
Bismuth Foundation

July 2018

Being converted for use in the pip package - DO NOT USE YET

"""


from tornado.tcpclient import TCPClient
import asyncio
import json
import time
import random


__version__ = '0.0.31B'


class AsyncClient():

    def __init__(self, server_list, app_log, loop, address=''):
        # print("async init", server_list)
        self.server_list = server_list
        self.app_log = app_log
        self.loop = loop
        self.address = address
        self.connected = False
        self.stream = None
        self._status = {"connected": False, "address": self.address}
        self.last_full_refresh = 0
        self.refreshing = False
        self.cached_aliases = {}
        self.needed_aliases = []
        self.ip_port = 'N/A'

    def status(self, address):
        self.address = address
        return self._status

    async def refresh(self):
        if not self.stream:
            return
        if self.refreshing:
            return
        try:
            self.refreshing = True
            statusget = await self._command("statusget")
            self._status['statusget'] = statusget
            self._status['status_version'] = statusget[7]
            self._status['stats_timestamp'] = statusget[9]

            # The two following ones only, depend on the address
            if self.address:
                stats_account = await self._command("balanceget", self.address)
                self._status['stats_account'] = stats_account
                self._status['address'] = self.address

                await self._send("addlistlim")
                await self._send(self.address)
                await self._send("20")
                self._status['addlist'] = await self._receive()

            self._status['block_get'] = await self._command("blocklast")
            self._status['diffget'] = await self._command("diffget")
            # Maybe not suitable to ask every 10 sec
            self._status['mpget'] = await self._command("mpget")

            if self.last_full_refresh < time.time() - 60 * 5:
                # Only run this if the info we got is older then 5 min old
                self._status['annverget'] = await self._command("annverget")
                self._status['annget'] = await self._command("annget")
                self.last_full_refresh = time.time()

            if len(self.needed_aliases):
                # Do we have aliases to resolve?
                solved = await self._command("aliasesget", self.needed_aliases)
                for index, alias in enumerate(self.needed_aliases):
                    self.cached_aliases[self.needed_aliases[index]] = solved[index]
                self.needed_aliases = []


        except Exception as e:
            self.app_log.error('refresh: {}'.format(e))
        finally:
            self.refreshing = False

    def aliases(self, needed_aliases):
        """
        cached_aliases hold the one we resolved already
        needed hold the one we, well, need
        This function has to return right away, can't wait for the results but will trigger anyway.
        :param needed_aliases:
        :return:
        """
        results = {}
        # willingly not using list/dict comprehensions to be easier to understand
        for alias in needed_aliases:
            if alias in self.cached_aliases:
                results[alias] = self.cached_aliases[alias]
            else:
                # display ... to say we are looking for it
                results[alias] = '[...] ' + alias
                if not alias in self.needed_aliases:
                    #  make sure they are unique
                    self.needed_aliases.append(alias)
        # they will be fetched in the background
        if len(self.needed_aliases):
            # force a refresh right away if we need aliases
            asyncio.run_coroutine_threadsafe(self.refresh(), self.loop)
        return results

    async def _receive(self):
        """
        Get a command, async version
        :param stream:
        :param ip:
        :return:
        """
        if self.stream:
            header = await self.stream.read_bytes(10)
            data_len = int(header)
            data = await self.stream.read_bytes(data_len)
            data = json.loads(data.decode("utf-8"))
            return data
        else:
            self.app_log.warning('receive: not connected')

    def receive(self, timeout=None):
        future = asyncio.run_coroutine_threadsafe(self._receive(), self.loop)
        return future.result(timeout)

    def send(self, data, timeout=None):
        future = asyncio.run_coroutine_threadsafe(self._send(data), self.loop)
        return future.result(timeout)

    def command(self, data,  param=None, timeout=None):
        future = asyncio.run_coroutine_threadsafe(self._command(data, param), self.loop)
        return future.result(timeout)

    async def _send(self, data):
        """
        sends an object to the stream, async.
        :param data:
        :param stream:
        :param ip:
        :return:
        """
        if self.stream:
            try:
                data = str(json.dumps(data))
                header = str(len(data)).encode("utf-8").zfill(10)
                full = header + data.encode('utf-8')
                await self.stream.write(full)
            except Exception as e:
                self.app_log.error("send_to_stream {} for ip {}".format(str(e), self.ip_port))
                self.stream = None
                self.connected = False
                raise
        else:
            self.app_log.warning('send: not connected')

    async def _command(self, data, param=None):
        if self.stream:
            await self._send(data)
            if param:
                await self._send(param)
            return await self._receive()
        else:
            self.app_log.warning('command: not connected')
            return None

    def convert_ip_port(self, ip, some_port):
        """
        Get ip and port, but extract port from ip if ip was as ip:port
        :param ip:
        :param some_port: default port
        :return: (ip, port)
        """
        if ':' in ip:
            ip, some_port = ip.split(':')
        return ip, some_port

    async def background(self):
        """
        background task that tries to stay connected all the time
        :return:
        """
        try:
            # Shuffle server list
            random.shuffle(self.server_list)
            try:
                # Force local host to be tried first, once.
                self.server_list.remove('127.0.0.1')
                self.server_list.insert(0, '127.0.0.1')
            except:
                pass

            while True:
                for self.ip_port in self.server_list:
                    self.app_log.warning("async client trying to connect to {}".format(self.ip_port))
                    try:
                        ip, port = self.convert_ip_port(self.ip_port, 5658)
                        self.stream = await TCPClient().connect(ip, port)
                        if self.stream:
                            self.connected = True
                            self.app_log.warning("connected to {}".format(self.ip_port))
                        while self.stream:
                            await self.refresh()
                            await asyncio.sleep(10)
                    except Exception as e:
                        self.app_log.error("Error in Stream: {} for {}".format(e, self.ip_port))
                    finally:
                        if self.stream:
                            self.stream.close()
                            self.stream = None
                        self.connected = False
                    await asyncio.sleep(5)
        except Exception as e:
            self.app_log.error("Error background {} for {}".format(e, self.ip_port))

connection = None

if __name__ == "__main__":
    print("I'm a module, can't run")
